###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import snappy

from autobahn.websocket.compress_base import PerMessageCompressOffer, \
    PerMessageCompressOfferAccept, \
    PerMessageCompressResponse, \
    PerMessageCompressResponseAccept, \
    PerMessageCompress

__all__ = (
    'PerMessageSnappyMixin',
    'PerMessageSnappyOffer',
    'PerMessageSnappyOfferAccept',
    'PerMessageSnappyResponse',
    'PerMessageSnappyResponseAccept',
    'PerMessageSnappy',
)


class PerMessageSnappyMixin:
    """
    Mixin class for this extension.
    """

    EXTENSION_NAME = "permessage-snappy"
    """
   Name of this WebSocket extension.
   """


class PerMessageSnappyOffer(PerMessageCompressOffer, PerMessageSnappyMixin):
    """
    Set of extension parameters for `permessage-snappy` WebSocket extension
    offered by a client to a server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension offer for `permessage-snappy` provided by a client to a server.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: object -- A new instance of :class:`autobahn.compress.PerMessageSnappyOffer`.
        """
        # extension parameter defaults
        ##
        acceptNoContextTakeover = False
        requestNoContextTakeover = False

        ##
        # verify/parse client ("client-to-server direction") parameters of permessage-snappy offer
        ##
        for p in params:

            if len(params[p]) > 1:
                raise Exception("multiple occurrence of extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

            val = params[p][0]

            if p == 'client_no_context_takeover':
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    acceptNoContextTakeover = True

            elif p == 'server_no_context_takeover':
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    requestNoContextTakeover = True

            else:
                raise Exception("illegal extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

        offer = cls(acceptNoContextTakeover,
                    requestNoContextTakeover)
        return offer

    def __init__(self,
                 acceptNoContextTakeover=True,
                 requestNoContextTakeover=False):
        """
        Constructor.

        :param acceptNoContextTakeover: Iff true, client accepts "no context takeover" feature.
        :type acceptNoContextTakeover: bool
        :param requestNoContextTakeover: Iff true, client request "no context takeover" feature.
        :type requestNoContextTakeover: bool
        """
        if type(acceptNoContextTakeover) != bool:
            raise Exception("invalid type %s for acceptNoContextTakeover" % type(acceptNoContextTakeover))

        self.acceptNoContextTakeover = acceptNoContextTakeover

        if type(requestNoContextTakeover) != bool:
            raise Exception("invalid type %s for requestNoContextTakeover" % type(requestNoContextTakeover))

        self.requestNoContextTakeover = requestNoContextTakeover

    def getExtensionString(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: str -- PMCE configuration string.
        """
        pmceString = self.EXTENSION_NAME
        if self.acceptNoContextTakeover:
            pmceString += "; client_no_context_takeover"
        if self.requestNoContextTakeover:
            pmceString += "; server_no_context_takeover"
        return pmceString

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'acceptNoContextTakeover': self.acceptNoContextTakeover,
                'requestNoContextTakeover': self.requestNoContextTakeover}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageSnappyOffer(acceptNoContextTakeover = %s, requestNoContextTakeover = %s)" % (self.acceptNoContextTakeover, self.requestNoContextTakeover)


class PerMessageSnappyOfferAccept(PerMessageCompressOfferAccept, PerMessageSnappyMixin):
    """
    Set of parameters with which to accept an `permessage-snappy` offer
    from a client by a server.
    """

    def __init__(self,
                 offer,
                 requestNoContextTakeover=False,
                 noContextTakeover=None):
        """
        Constructor.

        :param offer: The offer being accepted.
        :type offer: Instance of :class:`autobahn.compress.PerMessageSnappyOffer`.
        :param requestNoContextTakeover: Iff true, server request "no context takeover" feature.
        :type requestNoContextTakeover: bool
        :param noContextTakeover: Override server ("server-to-client direction") context takeover (this must be compatible with offer).
        :type noContextTakeover: bool
        """
        if not isinstance(offer, PerMessageSnappyOffer):
            raise Exception("invalid type %s for offer" % type(offer))

        self.offer = offer

        if type(requestNoContextTakeover) != bool:
            raise Exception("invalid type %s for requestNoContextTakeover" % type(requestNoContextTakeover))

        if requestNoContextTakeover and not offer.acceptNoContextTakeover:
            raise Exception("invalid value %s for requestNoContextTakeover - feature unsupported by client" % requestNoContextTakeover)

        self.requestNoContextTakeover = requestNoContextTakeover

        if noContextTakeover is not None:
            if type(noContextTakeover) != bool:
                raise Exception("invalid type %s for noContextTakeover" % type(noContextTakeover))

            if offer.requestNoContextTakeover and not noContextTakeover:
                raise Exception("invalid value %s for noContextTakeover - client requested feature" % noContextTakeover)

        self.noContextTakeover = noContextTakeover

    def getExtensionString(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: str -- PMCE configuration string.
        """
        pmceString = self.EXTENSION_NAME
        if self.offer.requestNoContextTakeover:
            pmceString += "; server_no_context_takeover"
        if self.requestNoContextTakeover:
            pmceString += "; client_no_context_takeover"
        return pmceString

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'offer': self.offer.__json__(),
                'requestNoContextTakeover': self.requestNoContextTakeover,
                'noContextTakeover': self.noContextTakeover}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageSnappyAccept(offer = %s, requestNoContextTakeover = %s, noContextTakeover = %s)" % (self.offer.__repr__(), self.requestNoContextTakeover, self.noContextTakeover)


class PerMessageSnappyResponse(PerMessageCompressResponse, PerMessageSnappyMixin):
    """
    Set of parameters for `permessage-snappy` responded by server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension response for `permessage-snappy` provided by a server to a client.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: object -- A new instance of :class:`autobahn.compress.PerMessageSnappyResponse`.
        """
        client_no_context_takeover = False
        server_no_context_takeover = False

        for p in params:

            if len(params[p]) > 1:
                raise Exception("multiple occurrence of extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

            val = params[p][0]

            if p == 'client_no_context_takeover':
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    client_no_context_takeover = True

            elif p == 'server_no_context_takeover':
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    server_no_context_takeover = True

            else:
                raise Exception("illegal extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

        response = cls(client_no_context_takeover,
                       server_no_context_takeover)
        return response

    def __init__(self,
                 client_no_context_takeover,
                 server_no_context_takeover):
        self.client_no_context_takeover = client_no_context_takeover
        self.server_no_context_takeover = server_no_context_takeover

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'client_no_context_takeover': self.client_no_context_takeover,
                'server_no_context_takeover': self.server_no_context_takeover}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageSnappyResponse(client_no_context_takeover = %s, server_no_context_takeover = %s)" % (self.client_no_context_takeover, self.server_no_context_takeover)


class PerMessageSnappyResponseAccept(PerMessageCompressResponseAccept, PerMessageSnappyMixin):
    """
    Set of parameters with which to accept an `permessage-snappy` response
    from a server by a client.
    """

    def __init__(self,
                 response,
                 noContextTakeover=None):
        """
        Constructor.

        :param response: The response being accepted.
        :type response: Instance of :class:`autobahn.compress.PerMessageSnappyResponse`.
        :param noContextTakeover: Override client ("client-to-server direction") context takeover (this must be compatible with response).
        :type noContextTakeover: bool
        """
        if not isinstance(response, PerMessageSnappyResponse):
            raise Exception("invalid type %s for response" % type(response))

        self.response = response

        if noContextTakeover is not None:
            if type(noContextTakeover) != bool:
                raise Exception("invalid type %s for noContextTakeover" % type(noContextTakeover))

            if response.client_no_context_takeover and not noContextTakeover:
                raise Exception("invalid value %s for noContextTakeover - server requested feature" % noContextTakeover)

        self.noContextTakeover = noContextTakeover

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'response': self.response.__json__(),
                'noContextTakeover': self.noContextTakeover}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageSnappyResponseAccept(response = %s, noContextTakeover = %s)" % (self.response.__repr__(), self.noContextTakeover)


class PerMessageSnappy(PerMessageCompress, PerMessageSnappyMixin):
    """
    `permessage-snappy` WebSocket extension processor.
    """

    @classmethod
    def createFromResponseAccept(cls, isServer, accept):
        pmce = cls(isServer,
                   accept.response.server_no_context_takeover,
                   accept.noContextTakeover if accept.noContextTakeover is not None else accept.response.client_no_context_takeover)
        return pmce

    @classmethod
    def createFromOfferAccept(cls, isServer, accept):
        pmce = cls(isServer,
                   accept.noContextTakeover if accept.noContextTakeover is not None else accept.offer.requestNoContextTakeover,
                   accept.requestNoContextTakeover)
        return pmce

    def __init__(self,
                 isServer,
                 server_no_context_takeover,
                 client_no_context_takeover):
        self._isServer = isServer
        self.server_no_context_takeover = server_no_context_takeover
        self.client_no_context_takeover = client_no_context_takeover

        self._compressor = None
        self._decompressor = None

    def __json__(self):
        return {'extension': self.EXTENSION_NAME,
                'server_no_context_takeover': self.server_no_context_takeover,
                'client_no_context_takeover': self.client_no_context_takeover}

    def __repr__(self):
        return "PerMessageSnappy(isServer = %s, server_no_context_takeover = %s, client_no_context_takeover = %s)" % (self._isServer, self.server_no_context_takeover, self.client_no_context_takeover)

    def startCompressMessage(self):
        if self._isServer:
            if self._compressor is None or self.server_no_context_takeover:
                self._compressor = snappy.StreamCompressor()
        else:
            if self._compressor is None or self.client_no_context_takeover:
                self._compressor = snappy.StreamCompressor()

    def compressMessageData(self, data):
        return self._compressor.add_chunk(data)

    def endCompressMessage(self):
        return ""

    def startDecompressMessage(self):
        if self._isServer:
            if self._decompressor is None or self.client_no_context_takeover:
                self._decompressor = snappy.StreamDecompressor()
        else:
            if self._decompressor is None or self.server_no_context_takeover:
                self._decompressor = snappy.StreamDecompressor()

    def decompressMessageData(self, data):
        return self._decompressor.decompress(data)

    def endDecompressMessage(self):
        pass



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import traceback

from autobahn.websocket import protocol
from autobahn.websocket import http
from autobahn.wamp.interfaces import ITransport
from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost

__all__ = ('WampWebSocketServerProtocol',
           'WampWebSocketClientProtocol',
           'WampWebSocketServerFactory',
           'WampWebSocketClientFactory')


class WampWebSocketProtocol:
    """
    Base class for WAMP-over-WebSocket transport mixins.
    """

    def _bailout(self, code, reason=None):
        if self.factory.debug_wamp:
            print("Failing WAMP-over-WebSocket transport: code = {0}, reason = '{1}'".format(code, reason))
        self.failConnection(code, reason)

    def onOpen(self):
        """
        Callback from :func:`autobahn.websocket.interfaces.IWebSocketChannel.onOpen`
        """
        # WebSocket connection established. Now let the user WAMP session factory
        # create a new WAMP session and fire off session open callback.
        try:
            self._session = self.factory._factory()
            self._session.onOpen(self)
        except Exception as e:
            if self.factory.debug_wamp:
                traceback.print_exc()
            # Exceptions raised in onOpen are fatal ..
            reason = "WAMP Internal Error ({0})".format(e)
            self._bailout(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_INTERNAL_ERROR, reason=reason)

    def onClose(self, wasClean, code, reason):
        """
        Callback from :func:`autobahn.websocket.interfaces.IWebSocketChannel.onClose`
        """
        # WAMP session might never have been established in the first place .. guard this!
        if hasattr(self, '_session') and self._session:
            # WebSocket connection lost - fire off the WAMP
            # session close callback
            # noinspection PyBroadException
            try:
                if self.factory.debug_wamp:
                    print("WAMP-over-WebSocket transport lost: wasClean = {0}, code = {1}, reason = '{2}'".format(wasClean, code, reason))
                self._session.onClose(wasClean)
            except Exception:
                # silently ignore exceptions raised here ..
                if self.factory.debug_wamp:
                    traceback.print_exc()
            self._session = None

    def onMessage(self, payload, isBinary):
        """
        Callback from :func:`autobahn.websocket.interfaces.IWebSocketChannel.onMessage`
        """
        try:
            for msg in self._serializer.unserialize(payload, isBinary):
                if self.factory.debug_wamp:
                    print("RX {0}".format(msg))
                self._session.onMessage(msg)

        except ProtocolError as e:
            if self.factory.debug_wamp:
                traceback.print_exc()
            reason = "WAMP Protocol Error ({0})".format(e)
            self._bailout(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_PROTOCOL_ERROR, reason=reason)

        except Exception as e:
            if self.factory.debug_wamp:
                traceback.print_exc()
            reason = "WAMP Internal Error ({0})".format(e)
            self._bailout(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_INTERNAL_ERROR, reason=reason)

    def send(self, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.send`
        """
        if self.isOpen():
            try:
                if self.factory.debug_wamp:
                    print("TX {0}".format(msg))
                payload, isBinary = self._serializer.serialize(msg)
            except Exception as e:
                # all exceptions raised from above should be serialization errors ..
                raise SerializationError("Unable to serialize WAMP application payload ({0})".format(e))
            else:
                self.sendMessage(payload, isBinary)
        else:
            raise TransportLost()

    def isOpen(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.isOpen`
        """
        return self._session is not None

    def close(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.close`
        """
        if self.isOpen():
            self.sendClose(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_NORMAL)
        else:
            raise TransportLost()

    def abort(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.abort`
        """
        if self.isOpen():
            self._bailout(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_GOING_AWAY)
        else:
            raise TransportLost()


ITransport.register(WampWebSocketProtocol)


def parseSubprotocolIdentifier(subprotocol):
    try:
        s = subprotocol.split('.')
        if s[0] != "wamp":
            raise Exception("invalid protocol %s" % s[0])
        version = int(s[1])
        serializerId = '.'.join(s[2:])
        return version, serializerId
    except:
        return None, None


class WampWebSocketServerProtocol(WampWebSocketProtocol):
    """
    Mixin for WAMP-over-WebSocket server transports.
    """

    STRICT_PROTOCOL_NEGOTIATION = True

    def onConnect(self, request):
        """
        Callback from :func:`autobahn.websocket.interfaces.IWebSocketChannel.onConnect`
        """
        headers = {}
        for subprotocol in request.protocols:
            version, serializerId = parseSubprotocolIdentifier(subprotocol)
            if version == 2 and serializerId in self.factory._serializers.keys():
                self._serializer = self.factory._serializers[serializerId]
                return subprotocol, headers

        if self.STRICT_PROTOCOL_NEGOTIATION:
            raise http.HttpException(http.BAD_REQUEST[0], "This server only speaks WebSocket subprotocols %s" % ', '.join(self.factory.protocols))
        else:
            # assume wamp.2.json
            self._serializer = self.factory._serializers['json']
            return None, headers


class WampWebSocketClientProtocol(WampWebSocketProtocol):
    """
    Mixin for WAMP-over-WebSocket client transports.
    """

    STRICT_PROTOCOL_NEGOTIATION = True

    def onConnect(self, response):
        """
        Callback from :func:`autobahn.websocket.interfaces.IWebSocketChannel.onConnect`
        """
        if response.protocol not in self.factory.protocols:
            if self.STRICT_PROTOCOL_NEGOTIATION:
                raise Exception("Server does not speak any of the WebSocket subprotocols we requested (%s)." % ', '.join(self.factory.protocols))
            else:
                # assume wamp.2.json
                serializerId = 'json'
        else:
            version, serializerId = parseSubprotocolIdentifier(response.protocol)

        self._serializer = self.factory._serializers[serializerId]


class WampWebSocketFactory:
    """
    Base class for WAMP-over-WebSocket transport factory mixins.
    """

    def __init__(self, factory, serializers=None, debug_wamp=False):
        """
        Ctor.

        :param factory: A callable that produces instances that implement
           :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable
        :param serializers: A list of WAMP serializers to use (or None for default
           serializers). Serializers must implement
           :class:`autobahn.wamp.interfaces.ISerializer`.
        :type serializers: list
        """
        assert(callable(factory))
        self._factory = factory

        self.debug_wamp = debug_wamp

        if serializers is None:
            serializers = []

            # try MsgPack WAMP serializer
            try:
                from autobahn.wamp.serializer import MsgPackSerializer
                serializers.append(MsgPackSerializer(batched=True))
                serializers.append(MsgPackSerializer())
            except ImportError:
                pass

            # try JSON WAMP serializer
            try:
                from autobahn.wamp.serializer import JsonSerializer
                serializers.append(JsonSerializer(batched=True))
                serializers.append(JsonSerializer())
            except ImportError:
                pass

            if not serializers:
                raise Exception("could not import any WAMP serializers")

        self._serializers = {}
        for ser in serializers:
            self._serializers[ser.SERIALIZER_ID] = ser

        self._protocols = ["wamp.2.%s" % ser.SERIALIZER_ID for ser in serializers]


class WampWebSocketServerFactory(WampWebSocketFactory):
    """
    Mixin for WAMP-over-WebSocket server transport factories.
    """


class WampWebSocketClientFactory(WampWebSocketFactory):
    """
    Mixin for WAMP-over-WebSocket client transport factories.
    """



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import time
import random
import sys
import re
import six
from datetime import datetime, timedelta
from pprint import pformat

if six.PY3:
    # Python 3
    # noinspection PyShadowingBuiltins
    xrange = range

__all__ = ("utcnow",
           "parseutc",
           "utcstr",
           "id",
           "newid",
           "rtime",
           "Stopwatch",
           "Tracker",
           "EqualityMixin")


def utcnow():
    """
    Get current time in UTC as ISO 8601 string.

    :returns: Current time as string in ISO 8601 format.
    :rtype: unicode
    """
    now = datetime.utcnow()
    return now.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


def utcstr(ts):
    """
    Format UTC timestamp in ISO 8601 format.

    :param ts: The timestamp to format.
    :type ts: instance of :py:class:`datetime.datetime`

    :returns: Timestamp formatted in ISO 8601 format.
    :rtype: unicode
    """
    if ts:
        return ts.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    else:
        return ts


def parseutc(datestr):
    """
    Parse an ISO 8601 combined date and time string, like i.e. ``"2011-11-23T12:23:00Z"``
    into a UTC datetime instance.

    .. deprecated:: 0.8.12
       Use the **iso8601** module instead (e.g. ``iso8601.parse_date("2014-05-23T13:03:44.123Z")``)

    :param datestr: The datetime string to parse.
    :type datestr: unicode

    :returns: The converted datetime object.
    :rtype: instance of :py:class:`datetime.datetime`
    """
    try:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        return None


# noinspection PyShadowingBuiltins
def id():
    """
    Generate a new random object ID from range **[0, 2**53]**.

    The upper bound **2**53** is chosen since it is the maximum integer that can be
    represented as a IEEE double such that all smaller integers are representable as well.

    Hence, IDs can be safely used with languages that use IEEE double as their
    main (or only) number type (JavaScript, Lua, etc).

    :returns: A random object ID.
    :rtype: int
    """
    # return random.randint(0, 9007199254740992) # this is what the WAMP spec says
    return random.randint(0, 2147483647)  # use a reduced ID space for now (2**31-1)


def newid(length=16):
    """
    Generate a new random object ID.

    :param length: The length (in chars) of the ID to generate.
    :type length: int

    :returns: A random object ID.
    :rtype: str
    """
    return ''.join([random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_") for _ in xrange(length)])


# Select the most precise walltime measurement function available
# on the platform
##
if sys.platform.startswith('win'):
    # On Windows, this function returns wall-clock seconds elapsed since the
    # first call to this function, as a floating point number, based on the
    # Win32 function QueryPerformanceCounter(). The resolution is typically
    # better than one microsecond
    _rtime = time.clock
    _ = _rtime()  # this starts wallclock
else:
    # On Unix-like platforms, this used the first available from this list:
    # (1) gettimeofday() -- resolution in microseconds
    # (2) ftime() -- resolution in milliseconds
    # (3) time() -- resolution in seconds
    _rtime = time.time


rtime = _rtime
"""
Precise wallclock time.

:returns: The current wallclock in seconds. Returned values are only guaranteed
   to be meaningful relative to each other.
:rtype: float
"""


class Stopwatch:
    """
    Stopwatch based on walltime.

    This can be used to do code timing and uses the most precise walltime measurement
    available on the platform. This is a very light-weight object,
    so create/dispose is very cheap.
    """

    def __init__(self, start=True):
        """
        :param start: If ``True``, immediately start the stopwatch.
        :type start: bool
        """
        self._elapsed = 0
        if start:
            self._started = rtime()
            self._running = True
        else:
            self._started = None
            self._running = False

    def elapsed(self):
        """
        Return total time elapsed in seconds during which the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if self._running:
            now = rtime()
            return self._elapsed + (now - self._started)
        else:
            return self._elapsed

    def pause(self):
        """
        Pauses the stopwatch and returns total time elapsed in seconds during which
        the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if self._running:
            now = rtime()
            self._elapsed += now - self._started
            self._running = False
            return self._elapsed
        else:
            return self._elapsed

    def resume(self):
        """
        Resumes a paused stopwatch and returns total elapsed time in seconds
        during which the stopwatch was running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        if not self._running:
            self._started = rtime()
            self._running = True
            return self._elapsed
        else:
            now = rtime()
            return self._elapsed + (now - self._started)

    def stop(self):
        """
        Stops the stopwatch and returns total time elapsed in seconds during which
        the stopwatch was (previously) running.

        :returns: The elapsed time in seconds.
        :rtype: float
        """
        elapsed = self.pause()
        self._elapsed = 0
        self._started = None
        self._running = False
        return elapsed


class Tracker:
    """
    A key-based statistics tracker.
    """

    def __init__(self, tracker, tracked):
        """
        """
        self.tracker = tracker
        self.tracked = tracked
        self._timings = {}
        self._offset = rtime()
        self._dt_offset = datetime.utcnow()

    def track(self, key):
        """
        Track elapsed for key.

        :param key: Key under which to track the timing.
        :type key: str
        """
        self._timings[key] = rtime()

    def diff(self, startKey, endKey, formatted=True):
        """
        Get elapsed difference between two previously tracked keys.

        :param startKey: First key for interval (older timestamp).
        :type startKey: str
        :param endKey: Second key for interval (younger timestamp).
        :type endKey: str
        :param formatted: If ``True``, format computed time period and return string.
        :type formatted: bool

        :returns: Computed time period in seconds (or formatted string).
        :rtype: float or str
        """
        if endKey in self._timings and startKey in self._timings:
            d = self._timings[endKey] - self._timings[startKey]
            if formatted:
                if d < 0.00001:  # 10us
                    s = "%d ns" % round(d * 1000000000.)
                elif d < 0.01:  # 10ms
                    s = "%d us" % round(d * 1000000.)
                elif d < 10:  # 10s
                    s = "%d ms" % round(d * 1000.)
                else:
                    s = "%d s" % round(d)
                return s.rjust(8)
            else:
                return d
        else:
            if formatted:
                return "n.a.".rjust(8)
            else:
                return None

    def absolute(self, key):
        """
        Return the UTC wall-clock time at which a tracked event occurred.

        :param key: The key
        :type key: str

        :returns: Timezone-naive datetime.
        :rtype: instance of :py:class:`datetime.datetime`
        """
        elapsed = self[key]
        if elapsed is None:
            raise KeyError("No such key \"%s\"." % elapsed)
        return self._dt_offset + timedelta(seconds=elapsed)

    def __getitem__(self, key):
        if key in self._timings:
            return self._timings[key] - self._offset
        else:
            return None

    def __iter__(self):
        return self._timings.__iter__()

    def __str__(self):
        return pformat(self._timings)


class EqualityMixin:
    """
    Mixing to add equality comparison operators to a class.

    Two objects are identical under this mixin, if and only if:

    1. both object have the same class
    2. all non-private object attributes are equal
    """

    def __eq__(self, other):
        """
        Compare this object to another object for equality.

        :param other: The other object to compare with.
        :type other: obj

        :returns: ``True`` iff the objects are equal.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        # we only want the actual message data attributes (not eg _serialize)
        for k in self.__dict__:
            if not k.startswith('_'):
                if not self.__dict__[k] == other.__dict__[k]:
                    return False
        return True
        # return (isinstance(other, self.__class__) and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        """
        Compare this object to another object for inequality.

        :param other: The other object to compare with.
        :type other: obj

        :returns: ``True`` iff the objects are not equal.
        :rtype: bool
        """
        return not self.__eq__(other)


def wildcards2patterns(wildcards):
    """
    Compute a list of regular expression patterns from a list of
    wildcard strings. A wildcard string uses '*' as a wildcard character
    matching anything.

    :param wildcards: List of wildcard strings to compute regular expression patterns for.
    :type wildcards: list of str
    :returns: Computed regular expressions.
    :rtype: list of obj
    """
    return [re.compile(wc.replace('.', '\.').replace('*', '.*')) for wc in wildcards]



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import sys

from autobahn.wamp import role
from autobahn.wamp import message
from autobahn.wamp.exception import ProtocolError

if sys.version_info < (2, 7):
    # noinspection PyUnresolvedReferences
    import unittest2 as unittest
else:
    # from twisted.trial import unittest
    import unittest


class Foo:
    pass


class TestIds(unittest.TestCase):

    def test_valid_ids(self):
        for val in [0, 1, 23, 42, 9007199254740992]:
            self.assertEqual(val, message.check_or_raise_id(val))

    def test_invalid_ids(self):
        for val in [-1, -9007199254740992, None, b"", b"abc", u"", u"abc", 0.9, Foo(), False, True, [], {}]:
            self.assertRaises(ProtocolError, message.check_or_raise_id, val)


class TestUris(unittest.TestCase):

    def test_valid_uris_loose_nonempty(self):
        for u in [u"com.myapp.topic1",
                  u"com.myapp.product.123",
                  u"com.myapp.product.1.delete",
                  u"Com-star.MyApp.**+$for",
                  u"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5",
                  u"hello\x24world",
                  u"hello\xC2\xA2world",
                  u"hello\xE2\x82\xACworld",
                  u"hello\xF0\xA4\xAD\xA2world",
                  ]:
            self.assertEqual(u, message.check_or_raise_uri(u))

    def test_invalid_uris_loose_nonempty(self):
        for u in [0,
                  None,
                  True,
                  False,
                  0.8,
                  b"abc",
                  Foo(),
                  u"",
                  u".",
                  u"com.",
                  u"com..product",
                  u"com.my app.product",
                  u"com.my\tapp.product",
                  u"com.my\napp.product",
                  u"com.myapp.product#",
                  u"com.#.product",
                  ]:
            self.assertRaises(ProtocolError, message.check_or_raise_uri, u)

    def test_valid_uris_loose_empty(self):
        for u in [u"com.myapp.topic1",
                  u"com.myapp..123",
                  u"com.myapp.product.1.",
                  u"com.",
                  u".",
                  u"",
                  u"Com-star.MyApp.**+$for..foo",
                  u"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5..foo",
                  u"hello\x24world..foo",
                  u"hello\xC2\xA2world..foo",
                  u"hello\xE2\x82\xACworld..foo",
                  u"hello\xF0\xA4\xAD\xA2world..foo",
                  ]:
            self.assertEqual(u, message.check_or_raise_uri(u, allowEmptyComponents=True))

    def test_invalid_uris_loose_empty(self):
        for u in [0,
                  None,
                  True,
                  False,
                  0.8,
                  b"abc",
                  Foo(),
                  u"com.my app.product",
                  u"com.my\tapp.product",
                  u"com.my\napp.product",
                  u"com.myapp.product#",
                  u"com.#.product",
                  ]:
            self.assertRaises(ProtocolError, message.check_or_raise_uri, u, allowEmptyComponents=True)

    def test_valid_uris_strict_nonempty(self):
        for u in [u"com.myapp.topic1",
                  u"com.myapp.product.123",
                  u"com.myapp.product.1.delete",
                  ]:
            self.assertEqual(u, message.check_or_raise_uri(u, strict=True))

    def test_invalid_uris_strict_nonempty(self):
        for u in [0,
                  None,
                  True,
                  False,
                  0.8,
                  b"abc",
                  Foo(),
                  u"",
                  u".",
                  u"com.",
                  u"com..product",
                  u"com.my app.product",
                  u"com.my\tapp.product",
                  u"com.my\napp.product",
                  u"com.myapp.product#",
                  u"com.#.product",
                  u"Com-star.MyApp.**+$for",
                  u"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5",
                  u"hello\x24world",
                  u"hello\xC2\xA2world",
                  u"hello\xE2\x82\xACworld",
                  u"hello\xF0\xA4\xAD\xA2world",
                  ]:
            self.assertRaises(ProtocolError, message.check_or_raise_uri, u, strict=True)

    def test_valid_uris_strict_empty(self):
        for u in [u"com.myapp.topic1",
                  u"com.myapp..123",
                  u"com.myapp.product.1.",
                  u"com.",
                  u".",
                  u"",
                  ]:
            self.assertEqual(u, message.check_or_raise_uri(u, strict=True, allowEmptyComponents=True))

    def test_invalid_uris_strict_empty(self):
        for u in [0,
                  None,
                  True,
                  False,
                  0.8,
                  b"abc",
                  Foo(),
                  u"com.my app.product",
                  u"com.my\tapp.product",
                  u"com.my\napp.product",
                  u"com.myapp.product#",
                  u"com.#.product",
                  u"Com-star.MyApp.**+$for..foo",
                  u"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5..foo",
                  u"hello\x24world..foo",
                  u"hello\xC2\xA2world..foo",
                  u"hello\xE2\x82\xACworld..foo",
                  u"hello\xF0\xA4\xAD\xA2world..foo",
                  ]:
            self.assertRaises(ProtocolError, message.check_or_raise_uri, u, strict=True, allowEmptyComponents=True)


class TestErrorMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Error(message.Call.MESSAGE_TYPE, 123456, u'com.myapp.error1')
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Error.MESSAGE_TYPE)
        self.assertEqual(msg[1], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[2], 123456)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], u'com.myapp.error1')

        e = message.Error(message.Call.MESSAGE_TYPE, 123456, u'com.myapp.error1', args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 7)
        self.assertEqual(msg[0], message.Error.MESSAGE_TYPE)
        self.assertEqual(msg[1], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[2], 123456)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], u'com.myapp.error1')
        self.assertEqual(msg[5], [1, 2, 3])
        self.assertEqual(msg[6], {u'foo': 23, u'bar': u'hello'})

    def test_parse_and_marshal(self):
        wmsg = [message.Error.MESSAGE_TYPE, message.Call.MESSAGE_TYPE, 123456, {}, u'com.myapp.error1']
        msg = message.Error.parse(wmsg)
        self.assertIsInstance(msg, message.Error)
        self.assertEqual(msg.request_type, message.Call.MESSAGE_TYPE)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.error, u'com.myapp.error1')
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Error.MESSAGE_TYPE, message.Call.MESSAGE_TYPE, 123456, {}, u'com.myapp.error1', [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Error.parse(wmsg)
        self.assertIsInstance(msg, message.Error)
        self.assertEqual(msg.request_type, message.Call.MESSAGE_TYPE)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.error, u'com.myapp.error1')
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.marshal(), wmsg)


class TestSubscribeMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Subscribe(123456, u'com.myapp.topic1')
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Subscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.topic1')

        e = message.Subscribe(123456, u'com.myapp.topic1', match=message.Subscribe.MATCH_PREFIX)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Subscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'match': u'prefix'})
        self.assertEqual(msg[3], u'com.myapp.topic1')

    def test_parse_and_marshal(self):
        wmsg = [message.Subscribe.MESSAGE_TYPE, 123456, {}, u'com.myapp.topic1']
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, u'com.myapp.topic1')
        self.assertEqual(msg.match, message.Subscribe.MATCH_EXACT)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Subscribe.MESSAGE_TYPE, 123456, {u'match': u'prefix'}, u'com.myapp.topic1']
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, u'com.myapp.topic1')
        self.assertEqual(msg.match, message.Subscribe.MATCH_PREFIX)
        self.assertEqual(msg.marshal(), wmsg)


class TestSubscribedMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Subscribed(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Subscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Subscribed.MESSAGE_TYPE, 123456, 789123]
        msg = message.Subscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribed)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.subscription, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnsubscribeMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Unsubscribe(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unsubscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Unsubscribe.MESSAGE_TYPE, 123456, 789123]
        msg = message.Unsubscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.subscription, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnsubscribedMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Unsubscribed(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 2)
        self.assertEqual(msg[0], message.Unsubscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)

    def test_parse_and_marshal(self):
        wmsg = [message.Unsubscribed.MESSAGE_TYPE, 123456]
        msg = message.Unsubscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribed)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.marshal(), wmsg)


class TestPublishMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Publish(123456, u'com.myapp.topic1')
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.topic1')

        e = message.Publish(123456, u'com.myapp.topic1', args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.topic1')
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {u'foo': 23, u'bar': u'hello'})

        e = message.Publish(123456, u'com.myapp.topic1', excludeMe=False, exclude=[300], eligible=[100, 200, 300], discloseMe=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'exclude_me': False, u'disclose_me': True, u'exclude': [300], u'eligible': [100, 200, 300]})
        self.assertEqual(msg[3], u'com.myapp.topic1')

    def test_parse_and_marshal(self):
        wmsg = [message.Publish.MESSAGE_TYPE, 123456, {}, u'com.myapp.topic1']
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, u'com.myapp.topic1')
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.excludeMe, None)
        self.assertEqual(msg.exclude, None)
        self.assertEqual(msg.eligible, None)
        self.assertEqual(msg.discloseMe, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Publish.MESSAGE_TYPE, 123456, {}, u'com.myapp.topic1', [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, u'com.myapp.topic1')
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.excludeMe, None)
        self.assertEqual(msg.exclude, None)
        self.assertEqual(msg.eligible, None)
        self.assertEqual(msg.discloseMe, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Publish.MESSAGE_TYPE, 123456, {u'exclude_me': False, u'disclose_me': True, u'exclude': [300], u'eligible': [100, 200, 300]}, u'com.myapp.topic1']
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, u'com.myapp.topic1')
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.excludeMe, False)
        self.assertEqual(msg.exclude, [300])
        self.assertEqual(msg.eligible, [100, 200, 300])
        self.assertEqual(msg.discloseMe, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestPublishedMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Published(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Published.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Published.MESSAGE_TYPE, 123456, 789123]
        msg = message.Published.parse(wmsg)
        self.assertIsInstance(msg, message.Published)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestEventMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Event(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})

        e = message.Event(123456, 789123, args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {u'foo': 23, u'bar': u'hello'})

        e = message.Event(123456, 789123, publisher=300)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {u'publisher': 300})

    def test_parse_and_marshal(self):
        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.publisher, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {}, [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.publisher, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {u'publisher': 300}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.publisher, 300)
        self.assertEqual(msg.marshal(), wmsg)


class TestRegisterMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Register(123456, u'com.myapp.procedure1')
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Register.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.procedure1')

        e = message.Register(123456, u'com.myapp.procedure1', pkeys=[10, 11, 12])
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Register.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'pkeys': [10, 11, 12]})
        self.assertEqual(msg[3], u'com.myapp.procedure1')

    def test_parse_and_marshal(self):
        wmsg = [message.Register.MESSAGE_TYPE, 123456, {}, u'com.myapp.procedure1']
        msg = message.Register.parse(wmsg)
        self.assertIsInstance(msg, message.Register)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, u'com.myapp.procedure1')
        self.assertEqual(msg.pkeys, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Register.MESSAGE_TYPE, 123456, {u'pkeys': [10, 11, 12]}, u'com.myapp.procedure1']
        msg = message.Register.parse(wmsg)
        self.assertIsInstance(msg, message.Register)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, u'com.myapp.procedure1')
        self.assertEqual(msg.pkeys, [10, 11, 12])
        self.assertEqual(msg.marshal(), wmsg)


class TestRegisteredMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Registered(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Registered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Registered.MESSAGE_TYPE, 123456, 789123]
        msg = message.Registered.parse(wmsg)
        self.assertIsInstance(msg, message.Registered)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnregisterMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Unregister(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unregister.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Unregister.MESSAGE_TYPE, 123456, 789123]
        msg = message.Unregister.parse(wmsg)
        self.assertIsInstance(msg, message.Unregister)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnregisteredMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Unregistered(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 2)
        self.assertEqual(msg[0], message.Unregistered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)

    def test_parse_and_marshal(self):
        wmsg = [message.Unregistered.MESSAGE_TYPE, 123456]
        msg = message.Unregistered.parse(wmsg)
        self.assertIsInstance(msg, message.Unregistered)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.marshal(), wmsg)


class TestCallMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Call(123456, u'com.myapp.procedure1')
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.procedure1')

        e = message.Call(123456, u'com.myapp.procedure1', args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], u'com.myapp.procedure1')
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {u'foo': 23, u'bar': u'hello'})

        e = message.Call(123456, u'com.myapp.procedure1', timeout=10000)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'timeout': 10000})
        self.assertEqual(msg[3], u'com.myapp.procedure1')

    def test_parse_and_marshal(self):
        wmsg = [message.Call.MESSAGE_TYPE, 123456, {}, u'com.myapp.procedure1']
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, u'com.myapp.procedure1')
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Call.MESSAGE_TYPE, 123456, {}, u'com.myapp.procedure1', [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, u'com.myapp.procedure1')
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Call.MESSAGE_TYPE, 123456, {u'timeout': 10000}, u'com.myapp.procedure1']
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, u'com.myapp.procedure1')
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, 10000)
        self.assertEqual(msg.marshal(), wmsg)


class TestCancelMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Cancel(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Cancel.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Cancel(123456, mode=message.Cancel.KILL)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Cancel.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'mode': message.Cancel.KILL})

    def test_parse_and_marshal(self):
        wmsg = [message.Cancel.MESSAGE_TYPE, 123456, {}]
        msg = message.Cancel.parse(wmsg)
        self.assertIsInstance(msg, message.Cancel)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Cancel.MESSAGE_TYPE, 123456, {u'mode': message.Cancel.KILL}]
        msg = message.Cancel.parse(wmsg)
        self.assertIsInstance(msg, message.Cancel)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, message.Cancel.KILL)
        self.assertEqual(msg.marshal(), wmsg)


class TestResultMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Result(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Result(123456, args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], [1, 2, 3])
        self.assertEqual(msg[4], {u'foo': 23, u'bar': u'hello'})

        e = message.Result(123456, progress=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'progress': True})

    def test_parse_and_marshal(self):
        wmsg = [message.Result.MESSAGE_TYPE, 123456, {}]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Result.MESSAGE_TYPE, 123456, {}, [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Result.MESSAGE_TYPE, 123456, {u'progress': True}]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestInvocationMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Invocation(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})

        e = message.Invocation(123456, 789123, args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {u'foo': 23, u'bar': u'hello'})

        e = message.Invocation(123456, 789123, timeout=10000)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {u'timeout': 10000})

    def test_parse_and_marshal(self):
        wmsg = [message.Invocation.MESSAGE_TYPE, 123456, 789123, {}]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Invocation.MESSAGE_TYPE, 123456, 789123, {}, [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Invocation.MESSAGE_TYPE, 123456, 789123, {u'timeout': 10000}]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, 10000)
        self.assertEqual(msg.marshal(), wmsg)


class TestInterruptMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Interrupt(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Interrupt.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Interrupt(123456, mode=message.Interrupt.KILL)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Interrupt.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'mode': message.Interrupt.KILL})

    def test_parse_and_marshal(self):
        wmsg = [message.Interrupt.MESSAGE_TYPE, 123456, {}]
        msg = message.Interrupt.parse(wmsg)
        self.assertIsInstance(msg, message.Interrupt)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Interrupt.MESSAGE_TYPE, 123456, {u'mode': message.Interrupt.KILL}]
        msg = message.Interrupt.parse(wmsg)
        self.assertIsInstance(msg, message.Interrupt)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, message.Interrupt.KILL)
        self.assertEqual(msg.marshal(), wmsg)


class TestYieldMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Yield(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Yield(123456, args=[1, 2, 3], kwargs={u'foo': 23, u'bar': u'hello'})
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], [1, 2, 3])
        self.assertEqual(msg[4], {u'foo': 23, u'bar': u'hello'})

        e = message.Yield(123456, progress=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {u'progress': True})

    def test_parse_and_marshal(self):
        wmsg = [message.Yield.MESSAGE_TYPE, 123456, {}]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Yield.MESSAGE_TYPE, 123456, {}, [1, 2, 3], {u'foo': 23, u'bar': u'hello'}]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {u'foo': 23, u'bar': u'hello'})
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Yield.MESSAGE_TYPE, 123456, {u'progress': True}]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestHelloMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Hello(u"realm1", [role.RoleBrokerFeatures()])
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Hello.MESSAGE_TYPE)
        self.assertEqual(msg[1], u"realm1")
        self.assertEqual(msg[2], {u'roles': {u'broker': {}}})

        e = message.Hello(u"realm1", [role.RoleBrokerFeatures(subscriber_blackwhite_listing=True)])
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Hello.MESSAGE_TYPE)
        self.assertEqual(msg[1], u"realm1")
        self.assertEqual(msg[2], {u'roles': {u'broker': {u'features': {u'subscriber_blackwhite_listing': True}}}})

    def test_parse_and_marshal(self):
        wmsg = [message.Hello.MESSAGE_TYPE, u"realm1", {u'roles': {u'broker': {}}}]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, u"realm1")
        self.assertEqual(msg.roles, [role.RoleBrokerFeatures()])
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Hello.MESSAGE_TYPE, u"realm1", {u'roles': {u'broker': {u'features': {u'subscriber_blackwhite_listing': True}}}}]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, u"realm1")
        self.assertEqual(msg.roles, [role.RoleBrokerFeatures(subscriber_blackwhite_listing=True)])
        self.assertEqual(msg.marshal(), wmsg)

    def test_str(self):
        e = message.Hello(u"realm1", [role.RoleBrokerFeatures()])
        self.assertIsInstance(str(e), str)


class TestGoodbyeMessage(unittest.TestCase):

    def test_ctor(self):
        reason = u'wamp.error.system_shutdown'
        reason_msg = u'The host is shutting down now.'

        e = message.Goodbye()
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {})
        self.assertEqual(msg[2], message.Goodbye.DEFAULT_REASON)

        e = message.Goodbye(reason=reason)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {})
        self.assertEqual(msg[2], reason)

        e = message.Goodbye(reason=reason, message=reason_msg)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {u'message': reason_msg})
        self.assertEqual(msg[2], reason)

    def test_parse_and_marshal(self):
        reason = u'wamp.error.system_shutdown'
        reason_msg = u'The host is shutting down now.'

        wmsg = [message.Goodbye.MESSAGE_TYPE]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, reason]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {u'message': 100}, reason]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {}, reason]
        msg = message.Goodbye.parse(wmsg)
        self.assertIsInstance(msg, message.Goodbye)
        self.assertEqual(msg.reason, reason)
        self.assertEqual(msg.message, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {u'message': reason_msg}, reason]
        msg = message.Goodbye.parse(wmsg)
        self.assertIsInstance(msg, message.Goodbye)
        self.assertEqual(msg.reason, reason)
        self.assertEqual(msg.message, reason_msg)
        self.assertEqual(msg.marshal(), wmsg)

    def test_str(self):
        e = message.Goodbye(reason=u'wamp.error.system_shutdown', message=u'The host is shutting down now.')
        self.assertIsInstance(str(e), str)


class TestHeartbeatMessage(unittest.TestCase):

    def test_ctor(self):
        e = message.Heartbeat(123, 456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Heartbeat.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123)
        self.assertEqual(msg[2], 456)

        e = message.Heartbeat(123, 456, u"discard me")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Heartbeat.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123)
        self.assertEqual(msg[2], 456)
        self.assertEqual(msg[3], u"discard me")

    def test_parse_and_marshal(self):
        wmsg = [message.Heartbeat.MESSAGE_TYPE]
        self.assertRaises(ProtocolError, message.Heartbeat.parse, wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 100]
        self.assertRaises(ProtocolError, message.Heartbeat.parse, wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 100, u"foo"]
        self.assertRaises(ProtocolError, message.Heartbeat.parse, wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 100, 0]
        self.assertRaises(ProtocolError, message.Heartbeat.parse, wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 100, 100, 100]
        self.assertRaises(ProtocolError, message.Heartbeat.parse, wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 123, 456]
        msg = message.Heartbeat.parse(wmsg)
        self.assertIsInstance(msg, message.Heartbeat)
        self.assertEqual(msg.incoming, 123)
        self.assertEqual(msg.outgoing, 456)
        self.assertEqual(msg.discard, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Heartbeat.MESSAGE_TYPE, 123, 456, u"discard me"]
        msg = message.Heartbeat.parse(wmsg)
        self.assertIsInstance(msg, message.Heartbeat)
        self.assertEqual(msg.incoming, 123)
        self.assertEqual(msg.outgoing, 456)
        self.assertEqual(msg.discard, u"discard me")
        self.assertEqual(msg.marshal(), wmsg)

    def test_str(self):
        e = message.Heartbeat(123, 456, u"discard me")
        self.assertIsInstance(str(e), str)


if __name__ == '__main__':
    unittest.main()



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import bz2

from autobahn.websocket.compress_base import PerMessageCompressOffer, \
    PerMessageCompressOfferAccept, \
    PerMessageCompressResponse, \
    PerMessageCompressResponseAccept, \
    PerMessageCompress

__all__ = (
    'PerMessageBzip2Mixin',
    'PerMessageBzip2Offer',
    'PerMessageBzip2OfferAccept',
    'PerMessageBzip2Response',
    'PerMessageBzip2ResponseAccept',
    'PerMessageBzip2',
)


class PerMessageBzip2Mixin:
    """
    Mixin class for this extension.
    """

    EXTENSION_NAME = "permessage-bzip2"
    """
   Name of this WebSocket extension.
   """

    COMPRESS_LEVEL_PERMISSIBLE_VALUES = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
   Permissible value for compression level parameter.
   """


class PerMessageBzip2Offer(PerMessageCompressOffer, PerMessageBzip2Mixin):
    """
    Set of extension parameters for `permessage-bzip2` WebSocket extension
    offered by a client to a server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension offer for `permessage-bzip2` provided by a client to a server.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: object -- A new instance of :class:`autobahn.compress.PerMessageBzip2Offer`.
        """
        # extension parameter defaults
        ##
        acceptMaxCompressLevel = False
        requestMaxCompressLevel = 0

        ##
        # verify/parse client ("client-to-server direction") parameters of permessage-bzip2 offer
        ##
        for p in params:

            if len(params[p]) > 1:
                raise Exception("multiple occurrence of extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

            val = params[p][0]

            if p == 'client_max_compress_level':
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    acceptMaxCompressLevel = True

            elif p == 'server_max_compress_level':
                try:
                    val = int(val)
                except:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    requestMaxCompressLevel = val

            else:
                raise Exception("illegal extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

        offer = cls(acceptMaxCompressLevel,
                    requestMaxCompressLevel)
        return offer

    def __init__(self,
                 acceptMaxCompressLevel=True,
                 requestMaxCompressLevel=0):
        """
        Constructor.

        :param acceptMaxCompressLevel: Iff true, client accepts "maximum compression level" parameter.
        :type acceptMaxCompressLevel: bool
        :param requestMaxCompressLevel: Iff non-zero, client requests given "maximum compression level" - must be 1-9.
        :type requestMaxCompressLevel: int
        """
        if type(acceptMaxCompressLevel) != bool:
            raise Exception("invalid type %s for acceptMaxCompressLevel" % type(acceptMaxCompressLevel))

        self.acceptMaxCompressLevel = acceptMaxCompressLevel

        if requestMaxCompressLevel != 0 and requestMaxCompressLevel not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
            raise Exception("invalid value %s for requestMaxCompressLevel - permissible values %s" % (requestMaxCompressLevel, self.COMPRESS_LEVEL_PERMISSIBLE_VALUES))

        self.requestMaxCompressLevel = requestMaxCompressLevel

    def getExtensionString(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: str -- PMCE configuration string.
        """
        pmceString = self.EXTENSION_NAME
        if self.acceptMaxCompressLevel:
            pmceString += "; client_max_compress_level"
        if self.requestMaxCompressLevel != 0:
            pmceString += "; server_max_compress_level=%d" % self.requestMaxCompressLevel
        return pmceString

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'acceptMaxCompressLevel': self.acceptMaxCompressLevel,
                'requestMaxCompressLevel': self.requestMaxCompressLevel}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageBzip2Offer(acceptMaxCompressLevel = %s, requestMaxCompressLevel = %s)" % (self.acceptMaxCompressLevel, self.requestMaxCompressLevel)


class PerMessageBzip2OfferAccept(PerMessageCompressOfferAccept, PerMessageBzip2Mixin):
    """
    Set of parameters with which to accept an `permessage-bzip2` offer
    from a client by a server.
    """

    def __init__(self,
                 offer,
                 requestMaxCompressLevel=0,
                 compressLevel=None):
        """
        Constructor.

        :param offer: The offer being accepted.
        :type offer: Instance of :class:`autobahn.compress.PerMessageBzip2Offer`.
        :param requestMaxCompressLevel: Iff non-zero, server requests given "maximum compression level" - must be 1-9.
        :type requestMaxCompressLevel: int
        :param compressLevel: Override server ("server-to-client direction") compress level (this must be compatible with offer).
        :type compressLevel: int
        """
        if not isinstance(offer, PerMessageBzip2Offer):
            raise Exception("invalid type %s for offer" % type(offer))

        self.offer = offer

        if requestMaxCompressLevel != 0 and requestMaxCompressLevel not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
            raise Exception("invalid value %s for requestMaxCompressLevel - permissible values %s" % (requestMaxCompressLevel, self.COMPRESS_LEVEL_PERMISSIBLE_VALUES))

        if requestMaxCompressLevel != 0 and not offer.acceptMaxCompressLevel:
            raise Exception("invalid value %s for requestMaxCompressLevel - feature unsupported by client" % requestMaxCompressLevel)

        self.requestMaxCompressLevel = requestMaxCompressLevel

        if compressLevel is not None:
            if compressLevel not in self.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                raise Exception("invalid value %s for compressLevel - permissible values %s" % (compressLevel, self.COMPRESS_LEVEL_PERMISSIBLE_VALUES))

            if offer.requestMaxCompressLevel != 0 and compressLevel > offer.requestMaxCompressLevel:
                raise Exception("invalid value %s for compressLevel - client requested lower maximum value" % compressLevel)

        self.compressLevel = compressLevel

    def getExtensionString(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: str -- PMCE configuration string.
        """
        pmceString = self.EXTENSION_NAME
        if self.offer.requestMaxCompressLevel != 0:
            pmceString += "; server_max_compress_level=%d" % self.offer.requestMaxCompressLevel
        if self.requestMaxCompressLevel != 0:
            pmceString += "; client_max_compress_level=%d" % self.requestMaxCompressLevel
        return pmceString

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'offer': self.offer.__json__(),
                'requestMaxCompressLevel': self.requestMaxCompressLevel,
                'compressLevel': self.compressLevel}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageBzip2Accept(offer = %s, requestMaxCompressLevel = %s, compressLevel = %s)" % (self.offer.__repr__(), self.requestMaxCompressLevel, self.compressLevel)


class PerMessageBzip2Response(PerMessageCompressResponse, PerMessageBzip2Mixin):
    """
    Set of parameters for `permessage-bzip2` responded by server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension response for `permessage-bzip2` provided by a server to a client.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: object -- A new instance of :class:`autobahn.compress.PerMessageBzip2Response`.
        """
        client_max_compress_level = 0
        server_max_compress_level = 0

        for p in params:

            if len(params[p]) > 1:
                raise Exception("multiple occurrence of extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

            val = params[p][0]

            if p == 'client_max_compress_level':
                try:
                    val = int(val)
                except:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    client_max_compress_level = val

            elif p == 'server_max_compress_level':
                try:
                    val = int(val)
                except:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                if val not in PerMessageBzip2Mixin.COMPRESS_LEVEL_PERMISSIBLE_VALUES:
                    raise Exception("illegal extension parameter value '%s' for parameter '%s' of extension '%s'" % (val, p, cls.EXTENSION_NAME))
                else:
                    server_max_compress_level = val

            else:
                raise Exception("illegal extension parameter '%s' for extension '%s'" % (p, cls.EXTENSION_NAME))

        response = cls(client_max_compress_level,
                       server_max_compress_level)
        return response

    def __init__(self,
                 client_max_compress_level,
                 server_max_compress_level):
        self.client_max_compress_level = client_max_compress_level
        self.server_max_compress_level = server_max_compress_level

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'client_max_compress_level': self.client_max_compress_level,
                'server_max_compress_level': self.server_max_compress_level}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageBzip2Response(client_max_compress_level = %s, server_max_compress_level = %s)" % (self.client_max_compress_level, self.server_max_compress_level)


class PerMessageBzip2ResponseAccept(PerMessageCompressResponseAccept, PerMessageBzip2Mixin):
    """
    Set of parameters with which to accept an `permessage-bzip2` response
    from a server by a client.
    """

    def __init__(self,
                 response,
                 compressLevel=None):
        """
        Constructor.

        :param response: The response being accepted.
        :type response: Instance of :class:`autobahn.compress.PerMessageBzip2Response`.
        :param compressLevel: Override client ("client-to-server direction") compress level (this must be compatible with response).
        :type compressLevel: int
        """
        if not isinstance(response, PerMessageBzip2Response):
            raise Exception("invalid type %s for response" % type(response))

        self.response = response

        if compressLevel is not None:
            if compressLevel not in self.WINDOW_SIZE_PERMISSIBLE_VALUES:
                raise Exception("invalid value %s for compressLevel - permissible values %s" % (compressLevel, self.COMPRESS_LEVEL_PERMISSIBLE_VALUES))

            if response.client_max_compress_level != 0 and compressLevel > response.client_max_compress_level:
                raise Exception("invalid value %s for compressLevel - server requested lower maximum value" % compressLevel)

        self.compressLevel = compressLevel

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: object -- JSON serializable representation.
        """
        return {'extension': self.EXTENSION_NAME,
                'response': self.response.__json__(),
                'compressLevel': self.compressLevel}

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: str -- Python string representation.
        """
        return "PerMessageBzip2ResponseAccept(response = %s, compressLevel = %s)" % (self.response.__repr__(), self.compressLevel)


class PerMessageBzip2(PerMessageCompress, PerMessageBzip2Mixin):
    """
    `permessage-bzip2` WebSocket extension processor.
    """
    DEFAULT_COMPRESS_LEVEL = 9

    @classmethod
    def createFromResponseAccept(cls, isServer, accept):
        pmce = cls(isServer,
                   accept.response.server_max_compress_level,
                   accept.compressLevel if accept.compressLevel is not None else accept.response.client_max_compress_level)
        return pmce

    @classmethod
    def createFromOfferAccept(cls, isServer, accept):
        pmce = cls(isServer,
                   accept.compressLevel if accept.compressLevel is not None else accept.offer.requestMaxCompressLevel,
                   accept.requestMaxCompressLevel)
        return pmce

    def __init__(self,
                 isServer,
                 server_max_compress_level,
                 client_max_compress_level):
        self._isServer = isServer
        self._compressor = None
        self._decompressor = None

        self.server_max_compress_level = server_max_compress_level if server_max_compress_level != 0 else self.DEFAULT_COMPRESS_LEVEL
        self.client_max_compress_level = client_max_compress_level if client_max_compress_level != 0 else self.DEFAULT_COMPRESS_LEVEL

    def __json__(self):
        return {'extension': self.EXTENSION_NAME,
                'isServer': self._isServer,
                'server_max_compress_level': self.server_max_compress_level,
                'client_max_compress_level': self.client_max_compress_level}

    def __repr__(self):
        return "PerMessageBzip2(isServer = %s, server_max_compress_level = %s, client_max_compress_level = %s)" % (self._isServer, self.server_max_compress_level, self.client_max_compress_level)

    def startCompressMessage(self):
        if self._isServer:
            if self._compressor is None:
                self._compressor = bz2.BZ2Compressor(self.server_max_compress_level)
        else:
            if self._compressor is None:
                self._compressor = bz2.BZ2Compressor(self.client_max_compress_level)

    def compressMessageData(self, data):
        return self._compressor.compress(data)

    def endCompressMessage(self):
        data = self._compressor.flush()

        # there seems to be no "flush without close stream", and after
        # full flush, compressor must not be reused
        self._compressor = None

        return data

    def startDecompressMessage(self):
        if self._decompressor is None:
            self._decompressor = bz2.BZ2Decompressor()

    def decompressMessageData(self, data):
        return self._decompressor.decompress(data)

    def endDecompressMessage(self):
        self._decompressor = None



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

from autobahn.wamp.uri import Pattern


def register(uri):
    """
    Decorator for WAMP procedure endpoints.
    """
    def decorate(f):
        assert(callable(f))
        if not hasattr(f, '_wampuris'):
            f._wampuris = []
        f._wampuris.append(Pattern(uri, Pattern.URI_TARGET_ENDPOINT))
        return f
    return decorate


def subscribe(uri):
    """
    Decorator for WAMP event handlers.
    """
    def decorate(f):
        assert(callable(f))
        if not hasattr(f, '_wampuris'):
            f._wampuris = []
        f._wampuris.append(Pattern(uri, Pattern.URI_TARGET_HANDLER))
        return f
    return decorate


def error(uri):
    """
    Decorator for WAMP error classes.
    """
    def decorate(cls):
        assert(issubclass(cls, Exception))
        if not hasattr(cls, '_wampuris'):
            cls._wampuris = []
        cls._wampuris.append(Pattern(uri, Pattern.URI_TARGET_EXCEPTION))
        return cls
    return decorate



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

from autobahn.wamp import error

__all__ = (
    'Error',
    'SessionNotReady',
    'SerializationError',
    'ProtocolError',
    'TransportLost',
    'ApplicationError',
    'NotAuthorized',
    'InvalidUri',
)


class Error(RuntimeError):
    """
    Base class for all exceptions related to WAMP.
    """
    def __init__(self, reason):
        """

        :param reason: Description of WAMP error that occurred (for logging purposes).
        :type reason: unicode
        """
        RuntimeError.__init__(self, reason)


class SessionNotReady(Error):
    """
    The application tried to perform a WAMP interaction, but the
    session is not yet fully established.
    """


class SerializationError(Error):
    """
    Exception raised when the WAMP serializer could not serialize the
    application payload (``args`` or ``kwargs`` for ``CALL``, ``PUBLISH``, etc).
    """


class ProtocolError(Error):
    """
    Exception raised when WAMP protocol was violated. Protocol errors
    are fatal and are handled by the WAMP implementation. They are
    not supposed to be handled at the application level.
    """


class TransportLost(Error):
    """
    Exception raised when the transport underlying the WAMP session
    was lost or is not connected.
    """
    def __init__(self):
        Error.__init__(self, u"WAMP transport lost")


class ApplicationError(Error):
    """
    Base class for all exceptions that can/may be handled
    at the application level.
    """

    INVALID_URI = u"wamp.error.invalid_uri"
    """
   Peer provided an incorrect URI for a URI-based attribute of a WAMP message
   such as a realm, topic or procedure.
   """

    NO_SUCH_PROCEDURE = u"wamp.error.no_such_procedure"
    """
   A Dealer could not perform a call, since not procedure is currently registered
   under the given URI.
   """

    PROCEDURE_ALREADY_EXISTS = u"wamp.error.procedure_already_exists"
    """
   A procedure could not be registered, since a procedure with the given URI is
   already registered.
   """

    NO_SUCH_REGISTRATION = u"wamp.error.no_such_registration"
    """
   A Dealer could not perform a unregister, since the given registration is not active.
   """

    NO_SUCH_SUBSCRIPTION = u"wamp.error.no_such_subscription"
    """
   A Broker could not perform a unsubscribe, since the given subscription is not active.
   """

    INVALID_ARGUMENT = u"wamp.error.invalid_argument"
    """
   A call failed, since the given argument types or values are not acceptable to the
   called procedure - in which case the *Callee* may throw this error. Or a Router
   performing *payload validation* checked the payload (``args`` / ``kwargs``) of a call,
   call result, call error or publish, and the payload did not conform.
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    SYSTEM_SHUTDOWN = u"wamp.error.system_shutdown"
    """
   The *Peer* is shutting down completely - used as a ``GOODBYE`` (or ``ABORT``) reason.
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    CLOSE_REALM = u"wamp.error.close_realm"
    """
   The *Peer* want to leave the realm - used as a ``GOODBYE`` reason.
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    GOODBYE_AND_OUT = u"wamp.error.goodbye_and_out"
    """
   A *Peer* acknowledges ending of a session - used as a ``GOOBYE`` reply reason.
   """

    NOT_AUTHORIZED = u"wamp.error.not_authorized"
    """
   A call, register, publish or subscribe failed, since the session is not authorized
   to perform the operation.
   """

    AUTHORIZATION_FAILED = u"wamp.error.authorization_failed"
    """
   A Dealer or Broker could not determine if the *Peer* is authorized to perform
   a join, call, register, publish or subscribe, since the authorization operation
   *itself* failed. E.g. a custom authorizer did run into an error.
   """

    NO_SUCH_REALM = u"wamp.error.no_such_realm"
    """
   Peer wanted to join a non-existing realm (and the *Router* did not allow to auto-create
   the realm).
   """

    NO_SUCH_ROLE = u"wamp.error.no_such_role"
    """
   A *Peer* was to be authenticated under a Role that does not (or no longer) exists on the Router.
   For example, the *Peer* was successfully authenticated, but the Role configured does not
   exists - hence there is some misconfiguration in the Router.
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    CANCELED = u"wamp.error.canceled"
    """
   A Dealer or Callee canceled a call previously issued (WAMP AP).
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    OPTION_DISALLOWED_DISCLOSE_ME = u"wamp.error.option_disallowed.disclose_me"
    """
   A Router rejected client request to disclose its identity (WAMP AP).
   """

    # FIXME: this currently isn't used neither in Autobahn nor Crossbar. Check!
    NO_ELIGIBLE_CALLEE = u"wamp.error.no_eligible_callee"
    """
   A *Dealer* could not perform a call, since a procedure with the given URI is registered,
   but *Callee Black- and Whitelisting* and/or *Caller Exclusion* lead to the
   exclusion of (any) *Callee* providing the procedure (WAMP AP).
   """

    def __init__(self, error, *args, **kwargs):
        """

        :param error: The URI of the error that occurred, e.g. ``wamp.error.not_authorized``.
        :type error: unicode
        """
        Exception.__init__(self, *args)
        self.kwargs = kwargs
        self.error = error

    def __str__(self):
        if self.kwargs and 'traceback' in self.kwargs:
            tb = ':\n' + '\n'.join(self.kwargs.pop('traceback')) + '\n'
            self.kwargs['traceback'] = '...'
        else:
            tb = ''
        return "ApplicationError('{0}', args = {1}, kwargs = {2}){3}".format(self.error, self.args, self.kwargs, tb)


@error(ApplicationError.NOT_AUTHORIZED)
class NotAuthorized(Exception):
    """
    Not authorized to perform the respective action.
    """


@error(ApplicationError.INVALID_URI)
class InvalidUri(Exception):
    """
    The URI for a topic, procedure or error is not a valid WAMP URI.
    """



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import sys
import inspect

from twisted.python import log
from twisted.application import service
from twisted.internet.defer import Deferred, \
    maybeDeferred, \
    DeferredList, \
    inlineCallbacks

from autobahn.wamp import protocol
from autobahn.wamp.types import ComponentConfig
from autobahn.websocket.protocol import parseWsUrl
from autobahn.twisted.websocket import WampWebSocketClientFactory

__all__ = (
    'FutureMixin',
    'ApplicationSession',
    'ApplicationSessionFactory',
    'ApplicationRunner',
    'Application',
    'Service'
)


class FutureMixin:
    """
    Mixin for Twisted style Futures ("Deferreds").
    """

    @staticmethod
    def _create_future():
        return Deferred()

    @staticmethod
    def _as_future(fun, *args, **kwargs):
        return maybeDeferred(fun, *args, **kwargs)

    @staticmethod
    def _resolve_future(future, value):
        future.callback(value)

    @staticmethod
    def _reject_future(future, value):
        future.errback(value)

    @staticmethod
    def _add_future_callbacks(future, callback, errback):
        return future.addCallbacks(callback, errback)

    @staticmethod
    def _gather_futures(futures, consume_exceptions=True):
        return DeferredList(futures, consumeErrors=consume_exceptions)


class ApplicationSession(FutureMixin, protocol.ApplicationSession):
    """
    WAMP application session for Twisted-based applications.
    """


class ApplicationSessionFactory(FutureMixin, protocol.ApplicationSessionFactory):
    """
    WAMP application session factory for Twisted-based applications.
    """

    session = ApplicationSession
    """
   The application session class this application session factory will use. Defaults to :class:`autobahn.twisted.wamp.ApplicationSession`.
   """


class ApplicationRunner:
    """
    This class is a convenience tool mainly for development and quick hosting
    of WAMP application components.

    It can host a WAMP application component in a WAMP-over-WebSocket client
    connecting to a WAMP router.
    """

    def __init__(self, url, realm, extra=None, debug=False, debug_wamp=False, debug_app=False):
        """

        :param url: The WebSocket URL of the WAMP router to connect to (e.g. `ws://somehost.com:8090/somepath`)
        :type url: unicode
        :param realm: The WAMP realm to join the application session to.
        :type realm: unicode
        :param extra: Optional extra configuration to forward to the application component.
        :type extra: dict
        :param debug: Turn on low-level debugging.
        :type debug: bool
        :param debug_wamp: Turn on WAMP-level debugging.
        :type debug_wamp: bool
        :param debug_app: Turn on app-level debugging.
        :type debug_app: bool
        """
        self.url = url
        self.realm = realm
        self.extra = extra or dict()
        self.debug = debug
        self.debug_wamp = debug_wamp
        self.debug_app = debug_app
        self.make = None

    def run(self, make, start_reactor=True):
        """
        Run the application component.

        :param make: A factory that produces instances of :class:`autobahn.asyncio.wamp.ApplicationSession`
           when called with an instance of :class:`autobahn.wamp.types.ComponentConfig`.
        :type make: callable
        """
        from twisted.internet import reactor

        isSecure, host, port, resource, path, params = parseWsUrl(self.url)

        # start logging to console
        if self.debug or self.debug_wamp or self.debug_app:
            log.startLogging(sys.stdout)

        # factory for use ApplicationSession
        def create():
            cfg = ComponentConfig(self.realm, self.extra)
            try:
                session = make(cfg)
            except Exception:
                # the app component could not be created .. fatal
                log.err()
                reactor.stop()
            else:
                session.debug_app = self.debug_app
                return session

        # create a WAMP-over-WebSocket transport client factory
        transport_factory = WampWebSocketClientFactory(create, url=self.url,
                                                       debug=self.debug, debug_wamp=self.debug_wamp)

        # start the client from a Twisted endpoint
        from twisted.internet.endpoints import clientFromString

        if isSecure:
            endpoint_descriptor = "ssl:{0}:{1}".format(host, port)
        else:
            endpoint_descriptor = "tcp:{0}:{1}".format(host, port)

        client = clientFromString(reactor, endpoint_descriptor)
        d = client.connect(transport_factory)

        # if an error happens on the connect(), we save the underlying
        # exception so that after the event-loop exits we can re-raise
        # it to the caller.

        class ErrorCollector:
            exception = None

            def __call__(self, failure):
                self.exception = failure.value
                # print(failure.getErrorMessage())
                reactor.stop()
        connect_error = ErrorCollector()
        d.addErrback(connect_error)

        # now enter the Twisted reactor loop
        if start_reactor:
            reactor.run()

        # if we exited due to a connection error, raise that to the
        # caller
        if connect_error.exception:
            raise connect_error.exception


class _ApplicationSession(ApplicationSession):
    """
    WAMP application session class used internally with :class:`autobahn.twisted.app.Application`.
    """

    def __init__(self, config, app):
        """

        :param config: The component configuration.
        :type config: Instance of :class:`autobahn.wamp.types.ComponentConfig`
        :param app: The application this session is for.
        :type app: Instance of :class:`autobahn.twisted.wamp.Application`.
        """
        # noinspection PyArgumentList
        ApplicationSession.__init__(self, config)
        self.app = app

    @inlineCallbacks
    def onConnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onConnect`
        """
        yield self.app._fire_signal('onconnect')
        self.join(self.config.realm)

    @inlineCallbacks
    def onJoin(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onJoin`
        """
        for uri, proc in self.app._procs:
            yield self.register(proc, uri)

        for uri, handler in self.app._handlers:
            yield self.subscribe(handler, uri)

        yield self.app._fire_signal('onjoined')

    @inlineCallbacks
    def onLeave(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onLeave`
        """
        yield self.app._fire_signal('onleave')
        self.disconnect()

    @inlineCallbacks
    def onDisconnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onDisconnect`
        """
        yield self.app._fire_signal('ondisconnect')


class Application:
    """
    A WAMP application. The application object provides a simple way of
    creating, debugging and running WAMP application components.
    """

    def __init__(self, prefix=None):
        """

        :param prefix: The application URI prefix to use for procedures and topics,
           e.g. ``"com.example.myapp"``.
        :type prefix: unicode
        """
        self._prefix = prefix

        # procedures to be registered once the app session has joined the router/realm
        self._procs = []

        # event handler to be subscribed once the app session has joined the router/realm
        self._handlers = []

        # app lifecycle signal handlers
        self._signals = {}

        # once an app session is connected, this will be here
        self.session = None

    def __call__(self, config):
        """
        Factory creating a WAMP application session for the application.

        :param config: Component configuration.
        :type config: Instance of :class:`autobahn.wamp.types.ComponentConfig`

        :returns: obj -- An object that derives of
           :class:`autobahn.twisted.wamp.ApplicationSession`
        """
        assert(self.session is None)
        self.session = _ApplicationSession(config, self)
        return self.session

    def run(self, url=u"ws://localhost:8080/ws", realm=u"realm1",
            debug=False, debug_wamp=False, debug_app=False,
            start_reactor=True):
        """
        Run the application.

        :param url: The URL of the WAMP router to connect to.
        :type url: unicode
        :param realm: The realm on the WAMP router to join.
        :type realm: unicode
        :param debug: Turn on low-level debugging.
        :type debug: bool
        :param debug_wamp: Turn on WAMP-level debugging.
        :type debug_wamp: bool
        :param debug_app: Turn on app-level debugging.
        :type debug_app: bool
        """
        runner = ApplicationRunner(url, realm,
                                   debug=debug, debug_wamp=debug_wamp, debug_app=debug_app)
        runner.run(self.__call__, start_reactor)

    def register(self, uri=None):
        """
        Decorator exposing a function as a remote callable procedure.

        The first argument of the decorator should be the URI of the procedure
        to register under.

        :Example:

        .. code-block:: python

           @app.register('com.myapp.add2')
           def add2(a, b):
              return a + b

        Above function can then be called remotely over WAMP using the URI `com.myapp.add2`
        the function was registered under.

        If no URI is given, the URI is constructed from the application URI prefix
        and the Python function name.

        :Example:

        .. code-block:: python

           app = Application('com.myapp')

           # implicit URI will be 'com.myapp.add2'
           @app.register()
           def add2(a, b):
              return a + b

        If the function `yields` (is a co-routine), the `@inlineCallbacks` decorator
        will be applied automatically to it. In that case, if you wish to return something,
        you should use `returnValue`:

        :Example:

        .. code-block:: python

           from twisted.internet.defer import returnValue

           @app.register('com.myapp.add2')
           def add2(a, b):
              res = yield stuff(a, b)
              returnValue(res)

        :param uri: The URI of the procedure to register under.
        :type uri: unicode
        """
        def decorator(func):
            if uri:
                _uri = uri
            else:
                assert(self._prefix is not None)
                _uri = "{0}.{1}".format(self._prefix, func.__name__)

            if inspect.isgeneratorfunction(func):
                func = inlineCallbacks(func)

            self._procs.append((_uri, func))
            return func
        return decorator

    def subscribe(self, uri=None):
        """
        Decorator attaching a function as an event handler.

        The first argument of the decorator should be the URI of the topic
        to subscribe to. If no URI is given, the URI is constructed from
        the application URI prefix and the Python function name.

        If the function yield, it will be assumed that it's an asynchronous
        process and inlineCallbacks will be applied to it.

        :Example:

        .. code-block:: python

           @app.subscribe('com.myapp.topic1')
           def onevent1(x, y):
              print("got event on topic1", x, y)

        :param uri: The URI of the topic to subscribe to.
        :type uri: unicode
        """
        def decorator(func):
            if uri:
                _uri = uri
            else:
                assert(self._prefix is not None)
                _uri = "{0}.{1}".format(self._prefix, func.__name__)

            if inspect.isgeneratorfunction(func):
                func = inlineCallbacks(func)

            self._handlers.append((_uri, func))
            return func
        return decorator

    def signal(self, name):
        """
        Decorator attaching a function as handler for application signals.

        Signals are local events triggered internally and exposed to the
        developer to be able to react to the application lifecycle.

        If the function yield, it will be assumed that it's an asynchronous
        coroutine and inlineCallbacks will be applied to it.

        Current signals :

           - `onjoined`: Triggered after the application session has joined the
              realm on the router and registered/subscribed all procedures
              and event handlers that were setup via decorators.
           - `onleave`: Triggered when the application session leaves the realm.

        .. code-block:: python

           @app.signal('onjoined')
           def _():
              # do after the app has join a realm

        :param name: The name of the signal to watch.
        :type name: unicode
        """
        def decorator(func):
            if inspect.isgeneratorfunction(func):
                func = inlineCallbacks(func)
            self._signals.setdefault(name, []).append(func)
            return func
        return decorator

    @inlineCallbacks
    def _fire_signal(self, name, *args, **kwargs):
        """
        Utility method to call all signal handlers for a given signal.

        :param name: The signal name.
        :type name: str
        """
        for handler in self._signals.get(name, []):
            try:
                # FIXME: what if the signal handler is not a coroutine?
                # Why run signal handlers synchronously?
                yield handler(*args, **kwargs)
            except Exception as e:
                # FIXME
                log.msg("Warning: exception in signal handler swallowed", e)


class Service(service.MultiService):
    """
    A WAMP application as a twisted service.
    The application object provides a simple way of creating, debugging and running WAMP application
    components inside a traditional twisted application

    This manages application lifecycle of the wamp connection using startService and stopService
    Using services also allows to create integration tests that properly terminates their connections

    It can host a WAMP application component in a WAMP-over-WebSocket client
    connecting to a WAMP router.
    """
    def __init__(self, url, realm, make, extra=None,
                 debug=False, debug_wamp=False, debug_app=False):
        """

        :param url: The WebSocket URL of the WAMP router to connect to (e.g. `ws://somehost.com:8090/somepath`)
        :type url: unicode
        :param realm: The WAMP realm to join the application session to.
        :type realm: unicode
        :param make: A factory that produces instances of :class:`autobahn.asyncio.wamp.ApplicationSession`
           when called with an instance of :class:`autobahn.wamp.types.ComponentConfig`.
        :type make: callable
        :param extra: Optional extra configuration to forward to the application component.
        :type extra: dict
        :param debug: Turn on low-level debugging.
        :type debug: bool
        :param debug_wamp: Turn on WAMP-level debugging.
        :type debug_wamp: bool
        :param debug_app: Turn on app-level debugging.
        :type debug_app: bool
        """
        self.url = url
        self.realm = realm
        self.extra = extra or dict()
        self.debug = debug
        self.debug_wamp = debug_wamp
        self.debug_app = debug_app
        self.make = make
        service.MultiService.__init__(self)
        self.setupService()

    def setupService(self):
        """
        Setup the application component.
        """
        isSecure, host, port, resource, path, params = parseWsUrl(self.url)

        # factory for use ApplicationSession
        def create():
            cfg = ComponentConfig(self.realm, self.extra)
            session = self.make(cfg)
            session.debug_app = self.debug_app
            return session

        # create a WAMP-over-WebSocket transport client factory
        transport_factory = WampWebSocketClientFactory(create, url=self.url,
                                                       debug=self.debug, debug_wamp=self.debug_wamp)

        # setup the client from a Twisted endpoint

        if isSecure:
            from twisted.application.internet import SSLClient
            clientClass = SSLClient
        else:
            from twisted.application.internet import TCPClient
            clientClass = TCPClient

        client = clientClass(host, port, transport_factory)
        client.setServiceParent(self)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import sys
import inspect

from autobahn import wamp
from autobahn.wamp.uri import Pattern

if sys.version_info < (2, 7):
    # noinspection PyUnresolvedReferences
    import unittest2 as unittest
else:
    # from twisted.trial import unittest
    import unittest


class TestUris(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_invalid_uris(self):
        for u in [u"",
                  u"com.myapp.<product:foo>.update",
                  u"com.myapp.<123:int>.update",
                  u"com.myapp.<:product>.update",
                  u"com.myapp.<product:>.update",
                  u"com.myapp.<int:>.update",
                  ]:
            self.assertRaises(Exception, Pattern, u, Pattern.URI_TARGET_ENDPOINT)

    def test_valid_uris(self):
        for u in [u"com.myapp.proc1",
                  u"123",
                  u"com.myapp.<product:int>.update",
                  ]:
            p = Pattern(u, Pattern.URI_TARGET_ENDPOINT)
            self.assertIsInstance(p, Pattern)

    def test_parse_uris(self):
        tests = [
            (u"com.myapp.<product:int>.update", [
                (u"com.myapp.0.update", {u'product': 0}),
                (u"com.myapp.123456.update", {u'product': 123456}),
                (u"com.myapp.aaa.update", None),
                (u"com.myapp..update", None),
                (u"com.myapp.0.delete", None),
            ]
            ),
            (u"com.myapp.<product:string>.update", [
                (u"com.myapp.box.update", {u'product': u'box'}),
                (u"com.myapp.123456.update", {u'product': u'123456'}),
                (u"com.myapp..update", None),
            ]
            )
        ]
        for test in tests:
            pat = Pattern(test[0], Pattern.URI_TARGET_ENDPOINT)
            for ptest in test[1]:
                uri = ptest[0]
                kwargs_should = ptest[1]
                if kwargs_should is not None:
                    args_is, kwargs_is = pat.match(uri)
                    self.assertEqual(kwargs_is, kwargs_should)
                else:
                    self.assertRaises(Exception, pat.match, uri)


class TestDecorators(unittest.TestCase):

    def test_decorate_endpoint(self):

        @wamp.register(u"com.calculator.square")
        def square(_):
            pass

        self.assertTrue(hasattr(square, '_wampuris'))
        self.assertTrue(type(square._wampuris) == list)
        self.assertEqual(len(square._wampuris), 1)
        self.assertIsInstance(square._wampuris[0], Pattern)
        self.assertTrue(square._wampuris[0].is_endpoint())
        self.assertFalse(square._wampuris[0].is_handler())
        self.assertFalse(square._wampuris[0].is_exception())
        self.assertEqual(square._wampuris[0].uri(), u"com.calculator.square")
        self.assertEqual(square._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        # noinspection PyUnusedLocal
        @wamp.register(u"com.myapp.product.<product:int>.update")
        def update_product(product=None, label=None):
            pass

        self.assertTrue(hasattr(update_product, '_wampuris'))
        self.assertTrue(type(update_product._wampuris) == list)
        self.assertEqual(len(update_product._wampuris), 1)
        self.assertIsInstance(update_product._wampuris[0], Pattern)
        self.assertTrue(update_product._wampuris[0].is_endpoint())
        self.assertFalse(update_product._wampuris[0].is_handler())
        self.assertFalse(update_product._wampuris[0].is_exception())
        self.assertEqual(update_product._wampuris[0].uri(), u"com.myapp.product.<product:int>.update")
        self.assertEqual(update_product._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        # noinspection PyUnusedLocal
        @wamp.register(u"com.myapp.<category:string>.<cid:int>.update")
        def update(category=None, cid=None):
            pass

        self.assertTrue(hasattr(update, '_wampuris'))
        self.assertTrue(type(update._wampuris) == list)
        self.assertEqual(len(update._wampuris), 1)
        self.assertIsInstance(update._wampuris[0], Pattern)
        self.assertTrue(update._wampuris[0].is_endpoint())
        self.assertFalse(update._wampuris[0].is_handler())
        self.assertFalse(update._wampuris[0].is_exception())
        self.assertEqual(update._wampuris[0].uri(), u"com.myapp.<category:string>.<cid:int>.update")
        self.assertEqual(update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

    def test_decorate_handler(self):

        @wamp.subscribe(u"com.myapp.on_shutdown")
        def on_shutdown():
            pass

        self.assertTrue(hasattr(on_shutdown, '_wampuris'))
        self.assertTrue(type(on_shutdown._wampuris) == list)
        self.assertEqual(len(on_shutdown._wampuris), 1)
        self.assertIsInstance(on_shutdown._wampuris[0], Pattern)
        self.assertFalse(on_shutdown._wampuris[0].is_endpoint())
        self.assertTrue(on_shutdown._wampuris[0].is_handler())
        self.assertFalse(on_shutdown._wampuris[0].is_exception())
        self.assertEqual(on_shutdown._wampuris[0].uri(), u"com.myapp.on_shutdown")
        self.assertEqual(on_shutdown._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        # noinspection PyUnusedLocal
        @wamp.subscribe(u"com.myapp.product.<product:int>.on_update")
        def on_product_update(product=None, label=None):
            pass

        self.assertTrue(hasattr(on_product_update, '_wampuris'))
        self.assertTrue(type(on_product_update._wampuris) == list)
        self.assertEqual(len(on_product_update._wampuris), 1)
        self.assertIsInstance(on_product_update._wampuris[0], Pattern)
        self.assertFalse(on_product_update._wampuris[0].is_endpoint())
        self.assertTrue(on_product_update._wampuris[0].is_handler())
        self.assertFalse(on_product_update._wampuris[0].is_exception())
        self.assertEqual(on_product_update._wampuris[0].uri(), u"com.myapp.product.<product:int>.on_update")
        self.assertEqual(on_product_update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.subscribe(u"com.myapp.<category:string>.<cid:int>.on_update")
        def on_update(category=None, cid=None, label=None):
            pass

        self.assertTrue(hasattr(on_update, '_wampuris'))
        self.assertTrue(type(on_update._wampuris) == list)
        self.assertEqual(len(on_update._wampuris), 1)
        self.assertIsInstance(on_update._wampuris[0], Pattern)
        self.assertFalse(on_update._wampuris[0].is_endpoint())
        self.assertTrue(on_update._wampuris[0].is_handler())
        self.assertFalse(on_update._wampuris[0].is_exception())
        self.assertEqual(on_update._wampuris[0].uri(), u"com.myapp.<category:string>.<cid:int>.on_update")
        self.assertEqual(on_update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

    def test_decorate_exception(self):

        @wamp.error(u"com.myapp.error")
        class AppError(Exception):
            pass

        self.assertTrue(hasattr(AppError, '_wampuris'))
        self.assertTrue(type(AppError._wampuris) == list)
        self.assertEqual(len(AppError._wampuris), 1)
        self.assertIsInstance(AppError._wampuris[0], Pattern)
        self.assertFalse(AppError._wampuris[0].is_endpoint())
        self.assertFalse(AppError._wampuris[0].is_handler())
        self.assertTrue(AppError._wampuris[0].is_exception())
        self.assertEqual(AppError._wampuris[0].uri(), u"com.myapp.error")
        self.assertEqual(AppError._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        @wamp.error(u"com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):
            pass

        self.assertTrue(hasattr(ProductInactiveError, '_wampuris'))
        self.assertTrue(type(ProductInactiveError._wampuris) == list)
        self.assertEqual(len(ProductInactiveError._wampuris), 1)
        self.assertIsInstance(ProductInactiveError._wampuris[0], Pattern)
        self.assertFalse(ProductInactiveError._wampuris[0].is_endpoint())
        self.assertFalse(ProductInactiveError._wampuris[0].is_handler())
        self.assertTrue(ProductInactiveError._wampuris[0].is_exception())
        self.assertEqual(ProductInactiveError._wampuris[0].uri(), u"com.myapp.product.<product:int>.product_inactive")
        self.assertEqual(ProductInactiveError._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.error(u"com.myapp.<category:string>.<product:int>.inactive")
        class ObjectInactiveError(Exception):
            pass

        self.assertTrue(hasattr(ObjectInactiveError, '_wampuris'))
        self.assertTrue(type(ObjectInactiveError._wampuris) == list)
        self.assertEqual(len(ObjectInactiveError._wampuris), 1)
        self.assertIsInstance(ObjectInactiveError._wampuris[0], Pattern)
        self.assertFalse(ObjectInactiveError._wampuris[0].is_endpoint())
        self.assertFalse(ObjectInactiveError._wampuris[0].is_handler())
        self.assertTrue(ObjectInactiveError._wampuris[0].is_exception())
        self.assertEqual(ObjectInactiveError._wampuris[0].uri(), u"com.myapp.<category:string>.<product:int>.inactive")
        self.assertEqual(ObjectInactiveError._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

    def test_match_decorated_endpoint(self):

        @wamp.register(u"com.calculator.square")
        def square(x):
            return x

        args, kwargs = square._wampuris[0].match(u"com.calculator.square")
        self.assertEqual(square(666, **kwargs), 666)

        @wamp.register(u"com.myapp.product.<product:int>.update")
        def update_product(product=None, label=None):
            return product, label

        args, kwargs = update_product._wampuris[0].match(u"com.myapp.product.123456.update")
        kwargs['label'] = "foobar"
        self.assertEqual(update_product(**kwargs), (123456, "foobar"))

        @wamp.register(u"com.myapp.<category:string>.<cid:int>.update")
        def update(category=None, cid=None, label=None):
            return category, cid, label

        args, kwargs = update._wampuris[0].match(u"com.myapp.product.123456.update")
        kwargs['label'] = "foobar"
        self.assertEqual(update(**kwargs), ("product", 123456, "foobar"))

    def test_match_decorated_handler(self):

        @wamp.subscribe(u"com.myapp.on_shutdown")
        def on_shutdown():
            pass

        args, kwargs = on_shutdown._wampuris[0].match(u"com.myapp.on_shutdown")
        self.assertEqual(on_shutdown(**kwargs), None)

        @wamp.subscribe(u"com.myapp.product.<product:int>.on_update")
        def on_product_update(product=None, label=None):
            return product, label

        args, kwargs = on_product_update._wampuris[0].match(u"com.myapp.product.123456.on_update")
        kwargs['label'] = "foobar"
        self.assertEqual(on_product_update(**kwargs), (123456, "foobar"))

        @wamp.subscribe(u"com.myapp.<category:string>.<cid:int>.on_update")
        def on_update(category=None, cid=None, label=None):
            return category, cid, label

        args, kwargs = on_update._wampuris[0].match(u"com.myapp.product.123456.on_update")
        kwargs['label'] = "foobar"
        self.assertEqual(on_update(**kwargs), ("product", 123456, "foobar"))

    def test_match_decorated_exception(self):

        @wamp.error(u"com.myapp.error")
        class AppError(Exception):

            def __init__(self, msg):
                Exception.__init__(self, msg)

            def __eq__(self, other):
                return self.__class__ == other.__class__ and \
                    self.args == other.args

        args, kwargs = AppError._wampuris[0].match(u"com.myapp.error")
        # noinspection PyArgumentList
        self.assertEqual(AppError(u"fuck", **kwargs), AppError(u"fuck"))

        @wamp.error(u"com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):

            def __init__(self, msg, product=None):
                Exception.__init__(self, msg)
                self.product = product

            def __eq__(self, other):
                return self.__class__ == other.__class__ and \
                    self.args == other.args and \
                    self.product == other.product

        args, kwargs = ProductInactiveError._wampuris[0].match(u"com.myapp.product.123456.product_inactive")
        self.assertEqual(ProductInactiveError("fuck", **kwargs), ProductInactiveError("fuck", 123456))

        @wamp.error(u"com.myapp.<category:string>.<product:int>.inactive")
        class ObjectInactiveError(Exception):

            def __init__(self, msg, category=None, product=None):
                Exception.__init__(self, msg)
                self.category = category
                self.product = product

            def __eq__(self, other):
                return self.__class__ == other.__class__ and \
                    self.args == other.args and \
                    self.category == other.category and \
                    self.product == other.product

        args, kwargs = ObjectInactiveError._wampuris[0].match(u"com.myapp.product.123456.inactive")
        self.assertEqual(ObjectInactiveError("fuck", **kwargs), ObjectInactiveError("fuck", "product", 123456))


class KwException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.kwargs = kwargs

# what if the WAMP error message received
# contains args/kwargs that cannot be
# consumed by the constructor of the exception
# class defined for the WAMP error URI?

# 1. we can bail out (but we are already signaling an error)
# 2. we can require a generic constructor
# 3. we can map only unconsumed args/kwargs to generic attributes
# 4. we can silently drop unconsumed args/kwargs


def getargs(fun):
    try:
        argspec = inspect.getargspec(fun)
    except:
        if fun == Exception.__init__:
            # `inspect.getargspec(Exception.__init__)` does work on PyPy, but not
            # on CPython, since `Exception.__init__` is C code in CPython that
            # cannot be reflected upon.
            argspec = inspect.ArgSpec(args=['self'], varargs='args', keywords=None, defaults=None)
        else:
            raise Exception("could not inspect function {0}".format(fun))

    args = argspec.args[:-len(argspec.defaults)]
    kwargs = argspec.args[-len(argspec.defaults):]

    return args, kwargs, argspec.varargs, argspec.keywords


class MockSession:

    def __init__(self):
        self._ecls_to_uri_pat = {}
        self._uri_to_ecls = {}

    def define(self, exception, error=None):
        if error is None:
            assert(hasattr(exception, '_wampuris'))
            self._ecls_to_uri_pat[exception] = exception._wampuris
            self._uri_to_ecls[exception._wampuris[0].uri()] = exception
        else:
            assert(not hasattr(exception, '_wampuris'))
            self._ecls_to_uri_pat[exception] = [Pattern(error, Pattern.URI_TARGET_HANDLER)]
            self._uri_to_ecls[error] = exception

    def map_error(self, error, args=None, kwargs=None):

        # FIXME:
        # 1. map to ecls based on error URI wildcard/prefix
        # 2. extract additional args/kwargs from error URI

        if error in self._uri_to_ecls:
            ecls = self._uri_to_ecls[error]
            try:
                # the following might fail, eg. TypeError when
                # signature of exception constructor is incompatible
                # with args/kwargs or when the exception constructor raises
                if kwargs:
                    if args:
                        exc = ecls(*args, **kwargs)
                    else:
                        exc = ecls(**kwargs)
                else:
                    if args:
                        exc = ecls(*args)
                    else:
                        exc = ecls()
            except Exception:
                # FIXME: log e
                exc = KwException(error, *args, **kwargs)
        else:
            # this never fails
            args = args or []
            kwargs = kwargs or {}
            exc = KwException(error, *args, **kwargs)
        return exc


class TestDecoratorsAdvanced(unittest.TestCase):

    def test_decorate_exception_non_exception(self):

        def test():
            # noinspection PyUnusedLocal
            @wamp.error(u"com.test.error")
            class Foo:
                pass

        self.assertRaises(Exception, test)

    def test_decorate_endpoint_multiple(self):

        # noinspection PyUnusedLocal
        @wamp.register(u"com.oldapp.oldproc")
        @wamp.register(u"com.calculator.square")
        def square(x):
            pass

        self.assertTrue(hasattr(square, '_wampuris'))
        self.assertTrue(type(square._wampuris) == list)
        self.assertEqual(len(square._wampuris), 2)

        for i in range(2):
            self.assertIsInstance(square._wampuris[i], Pattern)
            self.assertTrue(square._wampuris[i].is_endpoint())
            self.assertFalse(square._wampuris[i].is_handler())
            self.assertFalse(square._wampuris[i].is_exception())
            self.assertEqual(square._wampuris[i]._type, Pattern.URI_TYPE_EXACT)

        self.assertEqual(square._wampuris[0].uri(), u"com.calculator.square")
        self.assertEqual(square._wampuris[1].uri(), u"com.oldapp.oldproc")

    def test_marshal_decorated_exception(self):

        @wamp.error(u"com.myapp.error")
        class AppError(Exception):
            pass

        try:
            raise AppError("fuck")
        except Exception as e:
            self.assertEqual(e._wampuris[0].uri(), u"com.myapp.error")

        @wamp.error(u"com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):

            def __init__(self, msg, product=None):
                Exception.__init__(self, msg)
                self.product = product

        try:
            raise ProductInactiveError("fuck", 123456)
        except Exception as e:
            self.assertEqual(e._wampuris[0].uri(), u"com.myapp.product.<product:int>.product_inactive")

        session = MockSession()
        session.define(AppError)

    def test_define_exception_undecorated(self):

        session = MockSession()

        class AppError(Exception):
            pass

        # defining an undecorated exception requires
        # an URI to be provided
        self.assertRaises(Exception, session.define, AppError)

        session.define(AppError, u"com.myapp.error")

        exc = session.map_error(u"com.myapp.error")
        self.assertIsInstance(exc, AppError)

    def test_define_exception_decorated(self):

        session = MockSession()

        @wamp.error(u"com.myapp.error")
        class AppError(Exception):
            pass

        # when defining a decorated exception
        # an URI must not be provided
        self.assertRaises(Exception, session.define, AppError, u"com.myapp.error")

        session.define(AppError)

        exc = session.map_error(u"com.myapp.error")
        self.assertIsInstance(exc, AppError)

    def test_map_exception_undefined(self):

        session = MockSession()

        exc = session.map_error(u"com.myapp.error")
        self.assertIsInstance(exc, Exception)

    def test_map_exception_args(self):

        session = MockSession()

        @wamp.error(u"com.myapp.error")
        class AppError(Exception):
            pass

        @wamp.error(u"com.myapp.error.product_inactive")
        class ProductInactiveError(Exception):
            def __init__(self, product=None):
                self.product = product

        # define exceptions in mock session
        session.define(AppError)
        session.define(ProductInactiveError)

        for test in [
            # (u"com.myapp.foo.error", [], {}, KwException),
            (u"com.myapp.error", [], {}, AppError),
            (u"com.myapp.error", ["you are doing it wrong"], {}, AppError),
            (u"com.myapp.error", ["you are doing it wrong", 1, 2, 3], {}, AppError),

            (u"com.myapp.error.product_inactive", [], {}, ProductInactiveError),
            (u"com.myapp.error.product_inactive", [], {"product": 123456}, ProductInactiveError),
        ]:
            error, args, kwargs, ecls = test
            exc = session.map_error(error, args, kwargs)

            self.assertIsInstance(exc, ecls)
            self.assertEqual(list(exc.args), args)


if __name__ == '__main__':
    unittest.main()



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

try:
    import unittest2 as unittest
except ImportError:
    import unittest
from mock import patch


class FakeReactor:
    '''
    This just fakes out enough reactor methods so .run() can work.
    '''
    stop_called = False

    def __init__(self, to_raise):
        self.stop_called = False
        self.to_raise = to_raise

    def run(self, *args, **kw):
        raise self.to_raise

    def stop(self):
        self.stop_called = True

    def connectTCP(self, *args, **kw):
        raise RuntimeError("ConnectTCP shouldn't get called")


class TestWampTwistedRunner(unittest.TestCase):

    def test_connect_error(self):
        '''
        Ensure the runner doesn't swallow errors and that it exits the
        reactor properly if there is one.
        '''
        try:
            from autobahn.twisted.wamp import ApplicationRunner
            from twisted.internet.error import ConnectionRefusedError
            # the 'reactor' member doesn't exist until we import it
            from twisted.internet import reactor  # noqa: F401
        except ImportError:
            raise unittest.SkipTest('No twisted')

        runner = ApplicationRunner('ws://localhost:1', 'realm')
        exception = ConnectionRefusedError("It's a trap!")

        with patch('twisted.internet.reactor', FakeReactor(exception)) as mockreactor:
            self.assertRaises(
                ConnectionRefusedError,
                # pass a no-op session-creation method
                runner.run, lambda _: None, start_reactor=True
            )
            self.assertTrue(mockreactor.stop_called)

if __name__ == '__main__':
    unittest.main()
